import uuid
import time
import logging
from cookies import Cookie, Cookies

from aj.gate.gate import WorkerGate


class Session (object):
    """
    Holds the HTTP session data
    """

    def __init__(self, key, client_info={}):
        self.id = uuid.uuid4()
        self.key = key
        self.client_info = client_info
        self.data = {}
        self.identity = None
        self.touch()
        self.active = True
        self.gate = WorkerGate(self)
        self.gate.start()
        logging.debug('New session %s' % self.id)

    def destroy(self):
        logging.debug('Destroying session %s' % self.id)
        self.gate.stop()

    def deactivate(self):
        """
        Marks this session as dead
        """
        self.active = False

    def touch(self):
        """
        Updates the "last used" timestamp
        """
        self.timestamp = time.time()

    def is_dead(self):
        return not self.active or (time.time() - self.timestamp) > 3600

    def set_cookie(self, http_context):
        """
        Adds headers to :class:`aj.http.HttpContext` that set the session cookie
        """
        http_context.add_header('Set-Cookie', Cookie('session', self.key, path='/', httponly=True).render_response())




import gevent
import json
import logging
import re
import types

from aj.api import *


def url(pattern):
    """
    Exposes the decorated method of your :class:`HttpPlugin` via HTTP

    :param pattern: URL regex (no ``^`` and ``$`` required)
    :type  pattern: str
    :rtype: function

        Named capture groups will be fed to function as ``**kwargs``

    """

    def decorator(f):
        f._url_pattern = re.compile('^%s$' % pattern)
        return f
    return decorator


class BaseHttpHandler (object):
    """
    Base class for everything that can process HTTP requests
    """

    def handle(self, context):
        """
        Should create a HTTP response in the given ``context`` and return the plain output

        :param context: HTTP context
        :type  context: :class:`aj.http.HttpContext`
        """


@interface
class HttpPlugin (object):
    """
    A base plugin class for HTTP request handling::

        @plugin
        class TerminalHttp (BasePlugin, HttpPlugin):
            @url('/aj:terminal/(?P<id>\d+)')
            def get_page(self, context, id):
                if context.session.identity is None:
                    context.respond_redirect('/')
                context.add_header('Content-Type', 'text/html')
                context.respond_ok()
                return self.open_content('static/index.html').read()

    """

    def handle(self, context):
        """
        Finds and executes the handler for given request context (handlers are methods decorated with :func:`url` )

        :param context: HTTP context
        :type  context: :class:`aj.http.HttpContext`
        """

        for name, method in self.__class__.__dict__.iteritems():
            if hasattr(method, '_url_pattern'):
                method = getattr(self, name)
                match = method._url_pattern.match(context.path)
                if match:
                    context.route_data = match.groupdict()
                    data = method(context, **context.route_data)
                    if isinstance(data, types.GeneratorType):
                        return data
                    else:
                        return [data]


@interface
class SocketEndpoint (object):
    plugin = None

    def __init__(self, context):
        self.context = context
        self.greenlets = []

    def on_connect(self, message):
        pass

    def on_disconnect(self, message):
        pass

    def destroy(self):
        for gl in self.greenlets:
            gl.kill(block=False)

    def on_message(self, message, *args):
        pass

    def spawn(self, target, *args, **kwargs):
        logging.debug('Spawning sub-Socket Greenlet (in a namespace): %s' % target.__name__)
        greenlet = gevent.spawn(target, *args, **kwargs)
        self.greenlets.append(greenlet)
        return greenlet

    def send(self, data, plugin=None):
        self.context.worker.send_to_upstream({
            'type': 'socket',
            'message': {
                'plugin': plugin or self.plugin,
                'data': data,
            },
        })



import logging
import logging.handlers
import os
import sys
from termcolor import colored
from datetime import datetime

import aj

LOG_DIR = '/var/log/ajenti'
LOG_NAME = 'ajenti.log'
LOG_FILE = os.path.join(LOG_DIR, LOG_NAME)
LOG_PARAMS = {
    'master_pid': None,
    'tag': 'master',
}


class ConsoleHandler (logging.StreamHandler):
    def __init__(self, stream):
        logging.StreamHandler.__init__(self, stream)

    def handle(self, record):
        if not self.stream.isatty():
            return logging.StreamHandler.handle(self, record)

        s = ''
        d = datetime.fromtimestamp(record.created)
        s += colored(d.strftime("%d.%m.%Y %H:%M  "), 'white')

        process_tag = ''
        padding = ''
        if LOG_PARAMS['tag'] == 'master':
            if os.getpid() == LOG_PARAMS['master_pid']:
                process_tag = colored('master', 'yellow')
            else:
                LOG_PARAMS['tag'] = 'worker'
                #process_tag = colored('...   ', 'yellow')
        if LOG_PARAMS['tag'] == 'worker':
            process_tag = colored('worker', 'green')
            padding = '  '
        if LOG_PARAMS['tag'] == 'task':
            process_tag = colored('task  ', 'blue')
            padding = '    '
        s += colored('[', 'white')
        s += process_tag
        s += colored(' %5i]  ' % os.getpid(), 'white')

        if aj.debug:
            s += colored(('%15s:%-4s  ' % (record.filename[-15:], record.lineno)), 'grey', attrs=['bold'])

        l = ''
        if record.levelname == 'DEBUG':
            l = colored('DEBUG', 'white')
        if record.levelname == 'INFO':
            l = colored('INFO ', 'green', attrs=['bold'])
        if record.levelname == 'WARNING':
            l = colored('WARN ', 'yellow', attrs=['bold'])
        if record.levelname == 'ERROR':
            l = colored('ERROR', 'red', attrs=['bold'])
        s += l + '  '

        s += padding

        try:
            s += record.msg % record.args
        except:
            s += record.msg
        s += '\n'
        self.stream.write(s)


def init_console(log_level=logging.INFO):
    log = logging.getLogger()
    log.setLevel(logging.DEBUG)

    stdout = ConsoleHandler(sys.stdout)
    stdout.setLevel(log_level)

    dformatter = logging.Formatter('%(asctime)s %(levelname)-8s %(module)s.%(funcName)s(): %(message)s')
    stdout.setFormatter(dformatter)
    log.handlers = [stdout]


def init_log_directory():
    if not os.path.exists(LOG_DIR):
        os.mkdir(LOG_DIR)


def init_log_rotation():
    log = logging.getLogger()
    try:
        handler = logging.handlers.TimedRotatingFileHandler(
            os.path.join(LOG_DIR, LOG_NAME),
            when='midnight',
            backupCount=7
        )
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s: %(message)s'))
        log.addHandler(handler)
    except IOError:
        pass

    return log


def set_log_params(**kwargs):
    LOG_PARAMS.update(kwargs)



from datetime import timedelta


def str_fsize(sz):
    """
    Formats file size as string (i.e., 1.2 Mb)
    """
    if sz < 1024:
        return '%.1f bytes' % sz
    sz /= 1024.0
    if sz < 1024:
        return '%.1f KB' % sz
    sz /= 1024.0
    if sz < 1024:
        return '%.1f MB' % sz
    sz /= 1024.0
    if sz < 1024:
        return '%.1f GB' % sz
    sz /= 1024.0
    return '%.1f TB' % sz


def str_timedelta(s):
    """
    Formats a time delta (i.e., "5 days, 5:06:07")
    """
    return str(timedelta(0, s)).split('.')[0]



"""
Heavily based on https://github.com/Eugeny/gevent_openssl/blob/master/gevent_openssl/SSL.py
"""

import OpenSSL.SSL
import select
import socket
import sys


class SSLSocket(object):
    def __init__(self, context, sock=None):
        self._context = context
        self._socket = sock
        self._connection = OpenSSL.SSL.Connection(context, sock)
        self._connection.set_accept_state()
        self._makefile_refs = 0

    def __getattr__(self, attr):
        if attr not in ('_context', '_socket', '_connection', '_makefile_refs'):
            return getattr(self._connection, attr)
        else:
            raise AttributeError

    def __iowait(self, io_func, *args, **kwargs):
        timeout = self._socket.gettimeout() or 0.1
        fd = self._socket.fileno()
        while True:
            try:
                return io_func(*args, **kwargs)
            except (OpenSSL.SSL.WantReadError, OpenSSL.SSL.WantX509LookupError):
                sys.exc_clear()
                _, _, errors = select.select([fd], [], [fd], timeout)
                if errors:
                    break
            except OpenSSL.SSL.WantWriteError:
                sys.exc_clear()
                _, _, errors = select.select([], [fd], [fd], timeout)
                if errors:
                    break

    def accept(self):
        print 'accepting', self._connection, self._connection.accept
        _, _, _ = select.select([self._socket.fileno()], [], [])
        conn, addr = self._connection.accept()
        print '>', conn
        client = SSLSocket(self._context, connection=conn)
        return client, addr

    def connect(self, *args, **kwargs):
        return self.__iowait(self._connection.connect, *args, **kwargs)

    def send(self, data, flags=0):
        try:
            return self.__iowait(self._connection.send, data, flags)
        except OpenSSL.SSL.SysCallError as e:
            if e[0] == -1 and not data:
                # errors when writing empty strings are expected and can be ignored
                return 0
            raise

    def recv(self, bufsiz, flags=0):
        pending = self._connection.pending()
        if pending:
            return self._connection.recv(min(pending, bufsiz))
        try:
            return self.__iowait(self._connection.recv, bufsiz, flags)
        except OpenSSL.SSL.ZeroReturnError:
            return ''
        except OpenSSL.SSL.SysCallError as e:
            if e[0] == -1 and 'Unexpected EOF' in e[1]:
                # errors when reading empty strings are expected and can be ignored
                return ''
            raise

    def read(self, bufsiz, flags=0):
        return self.recv(bufsiz, flags)

    def write(self, buf, flags=0):
        return self.sendall(buf, flags)

    def close(self):
        if self._makefile_refs < 1:
            self._connection = None
            if self._socket:
                self._socket.close()
        else:
            self._makefile_refs -= 1

    def makefile(self, mode='r', bufsize=-1):
        self._makefile_refs += 1
        return socket._fileobject(self, mode, bufsize, close=True)



import os
import pwd
import stat
import yaml

from aj.api import *


class BaseConfig (object):
    def __init__(self):
        self.data = None

    def load(self):
        raise NotImplementedError()

    def save(self):
        raise NotImplementedError()


@service
class UserConfig (BaseConfig):
    def __init__(self, context):
        username = pwd.getpwuid(os.getuid())[0]
        dir = os.path.expanduser('~%s/.config' % username)
        if not os.path.exists(dir):
            os.makedirs(dir)
        self.path = os.path.join(dir, 'ajenti.yml')
        if os.path.exists(self.path):
            self.load()
        else:
            self.data = {}

    def load(self):
        self.data = yaml.load(open(self.path))

    def harden(self):
        os.chmod(self.path, stat.S_IRWXU)

    def save(self):
        with open(self.path, 'w') as f:
            f.write(yaml.safe_dump(self.data, default_flow_style=False, encoding='utf-8', allow_unicode=True))
        self.harden()



import logging
import subprocess
import os


# add subprocess.check_output to Python < 2.6
if not hasattr(subprocess, 'check_output'):
    def c_o(*args, **kwargs):
        kwargs['stdout'] = subprocess.PIPE
        popen = subprocess.Popen(*args, **kwargs)
        stdout, stderr = popen.communicate()
        return stdout
    subprocess.check_output = c_o


old_Popen = subprocess.Popen.__init__


def Popen(*args, **kwargs):
    logging.debug('Popen: %s' % (args[1],))
    __null = open(os.devnull, 'w')
    return old_Popen(
        stdin=kwargs.pop('stdin', subprocess.PIPE),
        stdout=kwargs.pop('stdout', __null),
        stderr=kwargs.pop('stderr', __null),
        *args, **kwargs)

subprocess.Popen.__init__ = Popen


# fix AttributeError
# a super-rude fix - DummyThread doesn't have a __block so provide an acquired one
"""import threading


def tbget(self):
    if hasattr(self, '__compat_lock'):
        return self.__compat_lock
    c = threading.Condition()
    c.acquire()
    return c


def tbset(self, l):
    self.__compat_lock = l


def tbdel(self):
    del self.__compat_lock

threading.Thread._Thread__block = property(tbget, tbset, tbdel)


# fix AttributeError("'Event' object has no attribute '_reset_internal_locks'",)
import threading
if not hasattr(threading.Event, '_reset_internal_locks'):
    def r_i_l(self):
        pass
    threading.Event._reset_internal_locks = r_i_l

"""
# suppress Requests logging
logging.getLogger("requests").setLevel(logging.WARNING)

# suppress simplejson
try:
    import simplejson
    _loads = simplejson.loads

    def wrap(fx):
        def f(*args, **kwargs):
            kwargs.pop('use_decimal', None)
            return fx(*args, **kwargs)
        return f

    simplejson.dumps = wrap(simplejson.dumps)
    simplejson.loads = wrap(simplejson.loads)
except:
    pass


# Suppress CORS headers in Socket.IO
from socketio.transports import BaseTransport

old_transport_init = BaseTransport.__init__

def new_transport_init(self, *args, **kwargs):
    old_transport_init(self, *args, **kwargs)
    self.headers = []

BaseTransport.__init__ = new_transport_init



# Re-add sslwrap to Python 2.7.9
"""
import inspect
import gevent.ssl
__ssl__ = __import__('ssl')

try:
    _ssl = __ssl__._ssl
except AttributeError:
    _ssl = __ssl__._ssl2


def new_sslwrap(sock, server_side=False, keyfile=None, certfile=None, cert_reqs=__ssl__.CERT_NONE, ssl_version=__ssl__.PROTOCOL_SSLv23, ca_certs=None, ciphers=None):
    context = __ssl__.SSLContext(ssl_version)
    context.verify_mode = cert_reqs or __ssl__.CERT_NONE
    if ca_certs:
        context.load_verify_locations(ca_certs)
    if certfile:
        context.load_cert_chain(certfile, keyfile)
    if ciphers:
        print ciphers
        context.set_ciphers(ciphers)

    caller_self = inspect.currentframe().f_back.f_locals['self']
    return context._wrap_socket(sock, server_side=server_side, ssl_sock=caller_self)

if not hasattr(_ssl, 'sslwrap') and not hasattr(gevent.ssl, 'SSLContext'):
    _ssl.sslwrap = new_sslwrap

"""


import gipc
import greenlet
import logging
import os
import signal

from aj.gate.stream import *
from aj.gate.worker import *
from aj.util import BroadcastQueue


class WorkerGate (object):
    def __init__(self, session):
        self.session = session
        self.process = None
        self.q_http_replies = BroadcastQueue()
        self.q_socket_messages = BroadcastQueue()

    def start(self):
        pipe_parent, pipe_child = gipc.pipe(duplex=True)
        self.stream = GateStreamServerEndpoint(pipe_parent)
        stream_child = GateStreamWorkerEndpoint(pipe_child)

        self.process = gipc.start_process(
            target=self._target,
            kwargs={
                'stream': stream_child,
                '_pipe': pipe_child,
            }
        )

        logging.debug('Started child process %s' % self.process.pid)

        self.stream_reader = gevent.spawn(self._stream_reader)

    def stop(self):
        logging.debug('Stopping child process %s' % self.process.pid)

        try:
            os.killpg(self.process.pid, signal.SIGTERM)
        except OSError as e:
            logging.debug('Child process %s is already dead: %s' % (self.process.pid, e))
            return

        self.process.terminate()
        self.process.join(0.125)
        try:
            os.kill(self.process.pid, 0)
            logging.debug('Child process %s did not stop, killing' % self.process.pid)
            os.kill(self.process.pid, signal.SIGKILL)
            os.killpg(self.process.pid, signal.SIGKILL)
            try:
                os.kill(self.process.pid, 0)
                logging.error('Child process %s did not stop after SIGKILL!' % self.process.pid)
            except OSError:
                pass
        except OSError:
            pass

        self.stream_reader.kill(block=False)

    def _stream_reader(self):
        try:
            while True:
                resp = self.stream.buffer_single_response(None)
                if not resp:
                    return
                self.stream.ack_response(resp.id)
                if resp.object['type'] == 'socket':
                    self.q_socket_messages.broadcast(resp)
                if resp.object['type'] == 'http':
                    self.q_http_replies.broadcast(resp)
                if resp.object['type'] == 'terminate':
                    self.session.deactivate()
        except greenlet.GreenletExit:
            pass

    def _target(self, stream=None, _pipe=None):
        self.worker = Worker(stream, self)
        self.worker.run()



import logging


class Context (object):
    def __init__(self, parent=None):
        self.parent = parent
        self.service_instances = {}
        self.component_instances = {}

    def __repr__(self):
        if self.parent:
            return '[Context %s (child of %s)]' % (id(self), id(self.parent))
        else:
            return '[Context %s]' % id(self)

    @staticmethod
    def get_fqdn(cls):
        return cls.__module__ + '.' + cls.__name__

    def get_service(self, cls):
        fqdn = self.get_fqdn(cls)
        if not fqdn in self.service_instances:
            self.service_instances[fqdn] = cls(self)
        return self.service_instances[fqdn]

    def get_components(self, cls):
        for component in cls._implementations: 
            fqdn = self.get_fqdn(component)
            if not fqdn in self.component_instances:
                self.component_instances[fqdn] = component(self)
            yield self.component_instances[fqdn]


def service(cls):
    if not cls:
        return None

    # Inject methods
    def get(cls, context):
        return context.get_service(cls)
    cls.get = get.__get__(cls)

    logging.debug('Registering [%s] (service)' % Context.get_fqdn(cls))

    return cls


def interface(cls):
    if not cls:
        return None

    cls._implementations = []

    # Inject methods
    def all(cls, context):
        return list(context.get_components(cls))
    cls.all = all.__get__(cls)

    def classes(cls):
        return list(cls._implementations)
    cls.classes = classes.__get__(cls)

    logging.debug('Registering [%s] (interface)' % Context.get_fqdn(cls))

    return cls


def component(interface):
    def decorator(cls):
        if not cls:
            return None

        # Run custom verificator if any
        if hasattr(cls, '__verify__'):
            if not cls.__verify__():
                return cls

        if not hasattr(interface, '_implementations'):
            logging.error('%s is not an @interface' % interface)

        logging.debug('Registering [%s] (implementation of [%s])' % (Context.get_fqdn(cls), Context.get_fqdn(interface)))
        interface._implementations.append(cls)
        return cls

    return decorator




import imp
import logging
import os
import subprocess
import sys
import traceback
import weakref
import yaml

import aj
from aj.api import *
from aj.util import *


@public
class PluginProvider (object):
    def provide(self):
        raise NotImplementedError()


@public
class DirectoryPluginProvider (PluginProvider):
    def __init__(self, path):
        self.path = os.path.abspath(path)

    def provide(self):
        found_plugins = []
        for dir in os.listdir(self.path):
            path = os.path.join(self.path, dir)
            if os.path.isdir(path):
                if os.path.exists(os.path.join(path, 'plugin.yml')):
                    found_plugins.append(path)
        return found_plugins


@public
class PythonPathPluginProvider (PluginProvider):
    def __init__(self):
        pass

    def provide(self):
        found_plugins = []
        for path in sys.path:
            if os.path.isdir(path):
                logging.debug('Looking for plugins in %s' % path)
                found_plugins += DirectoryPluginProvider(path).provide()
        return found_plugins


@public
class PluginLoadError (Exception):
    pass


@public
class PluginFormatError (PluginLoadError):
    def describe(self):
        return 'Plugin format error'

    def __str__(self):
        return 'format error'


@public
class PluginCrashed (PluginLoadError):
    def __init__(self, e):
        self.e = e
        self.traceback = traceback.format_exc()

    def describe(self):
        return 'Crashed: %s' % self.e

    def __str__(self):
        return 'crashed: %s' % self.e


@public
class Dependency (object):
    class Unsatisfied (PluginLoadError):
        def __init__(self):
            PluginLoadError.__init__(self, None)
            self.dependency = None

        def reason(self):
            pass

        def describe(self):
            return 'Dependency unsatisfied'

        def __str__(self):
            return '%s (%s)' % (self.dependency.__class__.__name__, self.reason())

    def satisfied(self):
        if hasattr(self, '_was_satisfied'):
            return self._was_satisfied
        self._was_satisfied = self.is_satisfied()
        return self._was_satisfied

    def build_exception(self):
        exception = self.Unsatisfied()
        exception.dependency = self
        return exception

    def check(self):
        if not self.satisfied():
            exception = self.build_exception()
            raise exception

    @property
    def value(self):
        return str(self)


@public
class ModuleDependency (Dependency):
    description = 'Python module'

    class Unsatisfied (Dependency.Unsatisfied):
        def reason(self):
            return '%s' % self.dependency.module_name

    def __init__(self, module_name=None):
        self.module_name = module_name

    def is_satisfied(self):
        if self.module_name in sys.modules:
            return True
        try:
            __import__(self.module_name)
            return True
        except:
            return False

    def __str__(self):
        return self.module_name


@public
class PluginDependency (Dependency):
    description = 'Plugin'

    class Unsatisfied (Dependency.Unsatisfied):
        def reason(self):
            return '%s' % self.dependency.plugin_name

    def __init__(self, plugin_name=None):
        self.plugin_name = plugin_name

    def is_satisfied(self):
        # get_order() only contains successfully loaded plugins
        return self.plugin_name in PluginManager.get(aj.context).get_order()

    def __str__(self):
        return self.plugin_name


@public
class BinaryDependency (Dependency):
    description = 'Application binary'

    class Unsatisfied (Dependency.Unsatisfied):
        def reason(self):
            return '%s' % self.dependency.binary_name

    def __init__(self, binary_name=None):
        self.binary_name = binary_name

    def is_satisfied(self):
        return subprocess.call(['which', self.binary_name]) == 0

    def __str__(self):
        return self.binary_name


@public
class FileDependency (Dependency):
    description = 'File'

    class Unsatisfied (Dependency.Unsatisfied):
        def reason(self):
            return '%s' % self.dependency.file_name

    def __init__(self, file_name=None):
        self.file_name = file_name

    def is_satisfied(self):
        return os.path.exists(self.file_name)

    def __str__(self):
        return self.file_name


@public
@service
class PluginManager (object):
    """
    Handles plugin loading and unloading
    """

    blacklist = []

    __plugins = {}
    __order = []
    __info = {}

    def __init__(self, context):
        self.context = context

    # Plugin loader
    def get_all(self):
        return self.__plugins

    def __getitem__(self, name):
        return self.__plugins[name]

    def get_content_path(self, name, path):
        return os.path.join(self[name].path, path)

    def get_order(self):
        return self.__order

    def load_all_from(self, providers):
        found = []
        for provider in providers:
            found += provider.provide()

        self.__info = {}
        for path in found:
            yml_info = yaml.load(open(os.path.join(path, 'plugin.yml')))
            self.__info[yml_info['name']] = {
                'info': yml_info,
                'path': path
            }

        for plugin, info in self.__info.iteritems():
            if not plugin in self.__plugins:
                self.load_recursive(plugin, info)

    def load_recursive(self, name, info):
        while True:
            try:
                return self.load(name, info)
            except PluginDependency.Unsatisfied as e:
                if e.dependency.plugin_name in self.get_all():
                    if self.get_all()[e.dependency.plugin_name].crash:
                        self.get_all()[name].crash = e
                        logging.warn(
                            'Plugin dependency unsatisfied: "%s" -> "%s"' % (name, e.dependency.plugin_name)
                        )
                        return
                try:
                    if e.dependency.plugin_name not in self.__info:
                        logging.warn(
                            'Plugin dependency unsatisfied: "%s" -> "%s"' % (name, e.dependency.plugin_name)
                        )
                        return
                    logging.debug('Preloading plugin dependency: "%s"' % e.dependency.plugin_name)
                    if not self.load_recursive(e.dependency.plugin_name, self.__info[e.dependency.plugin_name]):
                        self.get_all()[name].crash = e
                        return
                except:
                    raise

    def load(self, name, info):
        """
        Loads given plugin
        """
        logging.debug('Loading plugin "%s"' % name)
        try:
            plugin_info = PluginInfo(**info['info'])
            plugin_info.active = False
            plugin_info.name = name
            plugin_info.crash = None
            plugin_info.path = info['path']
            self.__plugins[name] = plugin_info

            for dependency in plugin_info.dependencies:
                dependency.check()

            plugin_info.active = True

            module_path, module_name = os.path.split(info['path'])

            try:
                mod = imp.load_module(
                    'aj.plugins.%s' % name,
                    *imp.find_module(module_name, [module_path])
                )
            except PluginFormatError:
                raise
            except Exception as e:
                raise PluginCrashed(e)

            if name in self.__order:
                self.__order.remove(name)
            self.__order.append(name)

            return True
        except PluginDependency.Unsatisfied as e:
            raise
        except PluginFormatError as e:
            logging.warn(' *** [%s] Plugin error: "%s"' % (name, e))
        except PluginCrashed as e:
            logging.warn(' *** [%s] Plugin crashed: "%s"' % (name, e))
            print(e.traceback)
            plugin_info.crash = e
        except Dependency.Unsatisfied as e:
            logging.debug(' *** [%s] skipping due to "%s"' % (name, e))
            plugin_info.crash = e
        except PluginLoadError as e:
            logging.warn(' *** [%s] Plugin failed to load: "%s"' % (name, e))
            plugin_info.crash = e




import grp
import json
import logging
import os
import pwd
import gevent
import traceback
import setproctitle
import sys

import aj
from aj.api import *
from aj.api.http import SocketEndpoint
from aj.http import HttpMiddlewareAggregator, HttpContext
from aj.auth import AuthenticationMiddleware
from aj.routing import CentralDispatcher
from aj.log import set_log_params


class WorkerSocketNamespace (object):
    def __init__(self, context, id):
        self.context = context
        self.id = id
        logging.debug('Socket namespace %s created' % self.id)
        self.endpoints = [cls(self.context) for cls in SocketEndpoint.classes()]

    def process_event(self, event, msg):
        for endpoint in self.endpoints:
            if not msg or msg['plugin'] in ['*', endpoint.plugin]:
                data = msg['data'] if msg else None
                try:
                    getattr(endpoint, 'on_%s' % event)(data)
                except Exception as e:
                    logging.error('Exception in socket event handler')
                    traceback.print_exc()

    def destroy(self):
        logging.debug('Socket namespace %s is being destroyed' % self.id)
        for endpoint in self.endpoints:
            endpoint.destroy()


class Worker (object):
    def __init__(self, stream, gate):
        self.stream = stream
        self.gate = gate
        aj.master = False
        os.setpgrp()
        setproctitle.setproctitle('%s session worker #%i' % (sys.argv[0], os.getpid()))
        set_log_params(tag='worker')

        logging.info('New worker #%s, EUID %s, EGID %s' % (
            os.getpid(), os.geteuid(), os.getegid(),
        ))

        self.context = Context(parent=aj.context)
        self.context.session = self.gate.session
        self.context.worker = self
        self.handler = HttpMiddlewareAggregator([
            AuthenticationMiddleware.get(self.context),
            CentralDispatcher.get(self.context),
        ])

        logging.info('Serving client %s' % (
            self.context.session.client_info['address'],
        ))

    def demote(self, username):
        if os.getuid() != 0:
            logging.warn('Running as a limited user, setuid() unavailable!')
            return

        uid = pwd.getpwnam(username).pw_uid
        gid = pwd.getpwnam(username).pw_gid

        logging.info('Worker %s is demoting to %s (UID %s / GID %s)' % (os.getpid(), username, uid, gid))

        groups = [g.gr_gid for g in grp.getgrall() if username in g.gr_mem or g.gr_gid == gid]
        os.setgroups(groups)
        os.setgid(gid)
        os.setuid(uid)
        logging.info('Demoted, EUID %s EGID %s' % (os.geteuid(), os.getegid()))

    def run(self):
        try:
            socket_namespaces = {}
            while True:
                rq = self.stream.recv()
                if not rq:
                    return
                if rq.object['type'] == 'http':
                    gevent.spawn(self.handle_http_request, rq)
                if rq.object['type'] == 'socket':
                    msg = rq.object['message']
                    nsid = rq.object['namespace']
                    event = rq.object['event']

                    if event == 'connect':
                        socket_namespaces[nsid] = WorkerSocketNamespace(self.context, nsid)

                    socket_namespaces[nsid].process_event(event, msg)

                    if event == 'disconnect':
                        socket_namespaces[nsid].destroy()
                        logging.debug('Socket disconnected, destroying endpoints')
        except Exception as e:
            logging.error('Worker crashed!')
            traceback.print_exc()

    def terminate(self):
        self.send_to_upstream({
            'type': 'terminate',
        })

    def handle_http_request(self, rq):
        response_object = {
            'type': 'http',
        }

        try:
            http_context = HttpContext.deserialize(rq.object['context'])

            # Generate response
            content = self.handler.handle(http_context)
            # ---
            response_object['content'] = list(content)
            response_object['status'] = http_context.status
            response_object['headers'] = http_context.headers
            self.stream.reply(rq, response_object)
        except Exception as e:
            traceback.print_exc()
            response_object.update({
                'error': str(e),
                'exception': repr(e),
            })
            self.stream.reply(rq, response_object)

    def send_to_upstream(self, object):
        self.stream.reply(None, object)


class WorkerError (Exception):
    def __init__(self, response):
        self.response = response

    def __str__(self):
        return json.dumps(self.response)



from aj.util import LazyModule
requests = LazyModule('requests')

import crypt
import json
import logging
import os
import pwd
import spwd
import subprocess

import aj
from aj.api import *
from aj.api.http import BaseHttpHandler
from aj.security.verifier import ClientCertificateVerificator
from aj.util import *


class SudoError (Exception):
    def __init__(self, message):
        self.message = message


@public
@service
class AuthenticationMiddleware (BaseHttpHandler):
    def __init__(self, context):
        self.context = context
        self.auth = AuthenticationService.get(self.context)
        if not hasattr(context, 'identity'):
            context.identity = None

    def handle(self, http_context):
        if aj.dev_autologin and not self.context.identity:
            username = pwd.getpwuid(os.geteuid()).pw_name
            self.auth.login(username, demote=os.geteuid() == 0)
        if http_context.env['SSL_CLIENT_VALID']:
            if not self.context.identity:
                username = http_context.env['SSL_CLIENT_USER']
                logging.info('SSL client certificate %s verified as %s' % (http_context.env['SSL_CLIENT_DIGEST'], username))
                try:
                    pwd.getpwnam(username)
                    found = True
                except KeyError:
                    found = False
                if found:
                    self.auth.login(username)

        http_context.add_header('X-Auth-Identity', str(self.context.identity or ''))


@public
@service
class AuthenticationService (BaseHttpHandler):
    def __init__(self, context):
        self.context = context

    def check_password(self, username, password):
        try:
            enc_pwd = spwd.getspnam(username)[1]

            if enc_pwd in ["NP", "!", "", None]:
                return "user '%s' has no password set" % user
            if enc_pwd in ["LK", "*"]:
                return "account is locked"
            if enc_pwd == "!!":
                return "password has expired"

            if crypt.crypt(password, enc_pwd) == enc_pwd:
                return True
            else:
                return False
        except KeyError:
            return False
        return False

    def check_sudo_password(self, username, password):
        sudo = subprocess.Popen(
            ['sudo', '-S', '-u', 'eugene', '--', 'sh', '-c', 'sudo -k; sudo -S echo'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        o, e = sudo.communicate(password + '\n')
        if sudo.returncode != 0:
            raise SudoError((o + e).splitlines()[-1].strip())
        return True

    def check_persona_assertion(self, assertion, audience):
        params = {
            'assertion': assertion,
            'audience': audience,
        }

        resp = requests.post('https://verifier.login.persona.org/verify', data=params, verify=True)

        if resp.ok:
            verification_data = json.loads(resp.content)
            if verification_data['status'] == 'okay':
                email = verification_data['email']
                return email
            else:
                raise Exception('Persona auth failed')
        else:
            raise Exception('Request failed')

    def client_certificate_callback(self, connection, x509, errno, depth, result):
        if depth == 0 and (errno == 9 or errno == 10):
            return False  # expired / not yet valid
        if not aj.config.data['ssl']['client_auth']['force']:
            return True
        user = ClientCertificateVerificator.get(aj.context).verify(x509)
        return bool(user)

    def get_identity(self):
        return self.context.identity

    def login(self, username, demote=True):
        logging.info('Authenticating session as %s' % username)
        if demote:
            self.context.worker.demote(username)
        self.context.identity = username



from __future__ import unicode_literals

from aj.util import LazyModule
psutil = LazyModule('psutil') # -2MB

import daemon
import locale
import logging
import os
import signal
import socket
import sys
import syslog
import ssl
import traceback

import aj
import aj.plugins
from aj.auth import AuthenticationService
from aj.http import HttpRoot, HttpMiddlewareAggregator
from aj.gate.middleware import GateMiddleware
from aj.plugins import *
from aj.api import *
from aj.util import make_report
from aj.util.sslsocket import SSLSocket
from aj.util.pidfile import PidFile
from aj.wsgi import RequestHandler

import gevent
import gevent.ssl
from gevent import monkey
from OpenSSL import SSL, crypto

# Gevent monkeypatch ---------------------
monkey.patch_all(select=True, thread=True, aggressive=False, subprocess=True)

from gevent.event import Event
import threading
threading.Event = Event
# ----------------------------------------

import aj.compat

from socketio.server import SocketIOServer


def run(config=None, plugin_providers=[], product_name='ajenti', dev_mode=False, debug_mode=False, autologin=False):
    if config is None:
        raise TypeError('`config` can\'t be None')

    reload(sys)
    sys.setdefaultencoding('utf8')

    aj.product = product_name
    aj.debug = debug_mode
    aj.dev = dev_mode
    aj.dev_autologin = autologin

    aj.init()
    aj.log.set_log_params(tag='master', master_pid=os.getpid())
    aj.context = Context()
    aj.config = config
    aj.plugin_providers = plugin_providers
    logging.info('Loading config from %s' % aj.config)
    aj.config.load()

    if aj.debug:
        logging.warn('Debug mode')
    if aj.dev:
        logging.warn('Dev mode')

    try:
        locale.setlocale(locale.LC_ALL, '')
    except:
        logging.warning('Couldn\'t set default locale')

    logging.info('Ajenti Core %s' % aj.version)
    logging.info('Detected platform: %s / %s' % (aj.platform, aj.platform_string))

    # Load plugins
    PluginManager.get(aj.context).load_all_from(aj.plugin_providers)
    if len(PluginManager.get(aj.context).get_all()) == 0:
        logging.warn('No plugins were loaded!')

    if aj.config.data['bind']['mode'] == 'unix':
        path = aj.config.data['bind']['socket']
        if os.path.exists(path):
            os.unlink(path)
        listener = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            listener.bind(path)
        except:
            logging.error('Could not bind to %s' % path)
            sys.exit(1)

    if aj.config.data['bind']['mode'] == 'tcp':
        host = aj.config.data['bind']['host']
        port = aj.config.data['bind']['port']
        listener = socket.socket(socket.AF_INET6 if ':' in host else socket.AF_INET, socket.SOCK_STREAM)
        if not aj.platform in ['freebsd', 'osx']:
            try:
                listener.setsockopt(socket.IPPROTO_TCP, socket.TCP_CORK, 1)
            except:
                logging.warn('Could not set TCP_CORK')
        listener.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        logging.info('Binding to [%s]:%s' % (host, port))
        try:
            listener.bind((host, port))
        except Exception, e:
            logging.error('Could not bind: %s' % str(e))
            sys.exit(1)

    # Fix stupid socketio bug (it tries to do *args[0][0])
    socket.socket.__getitem__ = lambda x, y: None

    listener.listen(10)

    gateway = GateMiddleware.get(aj.context)
    application = HttpRoot(HttpMiddlewareAggregator([gateway])).dispatch

    aj.server = SocketIOServer(
        listener,
        log=open(os.devnull, 'w'),
        application=application,
        handler_class=RequestHandler,
        policy_server=False,
        transports=[
            str('websocket'),
            str('flashsocket'),
            str('xhr-polling'),
            str('jsonp-polling'),
        ],
    )

    if aj.config.data['ssl']['enable'] and aj.config.data['bind']['mode'] == 'tcp':
        context = SSL.Context(SSL.TLSv1_METHOD)
        context.set_session_id(str(id(context)))
        context.set_options(SSL.OP_NO_SSLv2 | SSL.OP_NO_SSLv3)

        certificate = crypto.load_certificate(
            crypto.FILETYPE_PEM,
            open(aj.config.data['ssl']['certificate']).read()
        )
        private_key = crypto.load_privatekey(
            crypto.FILETYPE_PEM,
            open(aj.config.data['ssl']['certificate']).read()
        )

        context.use_certificate(certificate)
        context.use_privatekey(private_key)

        if aj.config.data['ssl']['client_auth']['enable']:
            # todo harden files
            logging.info('Enabling SSL client authentication')
            context.add_client_ca(certificate)
            context.get_cert_store().add_cert(certificate)
            verify_flags = SSL.VERIFY_PEER
            if aj.config.data['ssl']['client_auth']['force']:
                verify_flags |= SSL.VERIFY_FAIL_IF_NO_PEER_CERT
            context.set_verify(verify_flags, AuthenticationService.get(aj.context).client_certificate_callback)
            context.set_verify_depth(0)

        aj.server.ssl_args = {'server_side': True}
        aj.server.wrap_socket = lambda socket, **ssl: SSLSocket(context, socket)
        logging.info('SSL enabled')

    # auth.log
    try:
        syslog.openlog(
            ident=str(aj.product),
            facility=syslog.LOG_AUTH,
        )
    except:
        syslog.openlog(aj.product)

    def cleanup():
        if hasattr(cleanup, '_started'):
            return
        cleanup._started = True
        logging.info('Process %s exiting normally' % os.getpid())
        gevent.signal(signal.SIGINT, lambda: None)
        gevent.signal(signal.SIGTERM, lambda: None)
        if aj.master:
            gateway.destroy()

        p = psutil.Process(os.getpid())
        for c in p.get_children(recursive=True):
            try:
                os.killpg(c.pid, signal.SIGTERM)
                os.killpg(c.pid, signal.SIGKILL)
            except OSError:
                pass
        sys.exit(0)

    try:
        gevent.signal(signal.SIGINT, cleanup)
        gevent.signal(signal.SIGTERM, cleanup)
    except:
        pass

    aj.server.serve_forever()

    if not aj.master:
        # child process, server is stopped, wait until killed
        gevent.wait()
        #while True:
        #    gevent.sleep(3600)

    if hasattr(aj.server, 'restart_marker'):
        logging.warn('Restarting by request')
        cleanup()

        fd = 20  # Close all descriptors. Creepy thing
        while fd > 2:
            try:
                os.close(fd)
                logging.debug('Closed descriptor #%i' % fd)
            except:
                pass
            fd -= 1

        os.execv(sys.argv[0], sys.argv)
    else:
        if aj.master:
            logging.debug('Server stopped')


def handle_crash(exc):
    # todo rework this
    logging.error('Fatal crash occured')
    traceback.print_exc()
    exc.traceback = traceback.format_exc(exc)
    report_path = '/root/%s-crash.txt' % aj.product
    try:
        report = open(report_path, 'w')
    except:
        report_path = './%s-crash.txt' % aj.product
        report = open(report_path, 'w')
    report.write(make_report(exc))
    report.close()
    logging.error('Crash report written to %s' % report_path)
    #logging.error('Please submit it to https://github.com/Eugeny/ajenti/issues/new')


def start(daemonize=False, log_level=logging.INFO, **kwargs):
    if daemonize:
        aj.log.init_log_directory()
        logfile = open(aj.log.LOG_FILE, 'w+')
        context = daemon.DaemonContext(
            pidfile=PidFile('/var/run/ajenti.pid'),
            stdout=logfile,
            stderr=logfile,
            detach_process=True
        )
        with context:
            gevent.reinit()
            aj.log.init_log_rotation()
            try:
                run(**kwargs)
            except Exception as e:
                handle_crash(e)
    else:
        try:
            run(**kwargs)
        except KeyboardInterrupt:
            pass
        except Exception as e:
            handle_crash(e)




#!/usr/bin/env python
from setuptools import setup
from setuptools.command.install import install as _install
from setuptools import find_packages

import os
import socket
import subprocess

import aj


def _post_install (dir):
    config_path = '/etc/ajenti/config.yml'
    if not os.path.exists(config_path):
        open(os.path.exists(config_path), 'w').write('''
bind:
  host: 0.0.0.0
  mode: tcp
  port: 8000
color: default
emails:
name: %s
ssl:
  certificate:
  client_auth:
    certificates: []
    enable: false
    force: false
  enable: false
''' % socket.gethostname())
    subprocess.call(['ajenti-ssl-gen', socket.gethostname()])



class install (_install):
    def run(self):
        _install.run(self)
        self.execute(_post_install, [self.install_lib], msg='Running post install script')


setup(
    name='ajenti-panel',
    version=aj.__version__,
    install_requires=[
        'aj==%s' % aj.__version__,
        'pyyaml',
    ],
    description='Ajenti core based panel',
    author='Eugene Pankov',
    author_email='e@ajenti.org',
    url='http://ajenti.org/',
    packages=find_packages(),
    scripts=['ajenti-panel', 'ajenti-ssl-gen', 'ajenti-client-ssl-gen'],
    cmdclass={'install': install},
)



from aj.api.di import *


class PluginInfo:
    """
    Describes a loaded plugin package
    """

    def __init__(self, **kwargs):
        self.name = ''
        self.description = ''
        self.icon = None
        self.author = ''
        self.homepage = ''
        self.dependencies = []

        def init():
            pass

        self.init = init

        for k in kwargs:
            setattr(self, k, kwargs[k])

    def __cmp__(self, other):
        return cmp(self.name, other.name)

    def __hash__(self):
        return hash(self.name)


__all__ = [
    'PluginInfo',
    'Context',
    'component',
    'interface',
    'service',
]



import os
import fcntl


class PidFile(object):
    """
    Context manager that locks a pid file.  Implemented as class
    not generator because daemon.py is calling .__exit__() with no parameters
    instead of the None, None, None specified by PEP-343.
    """

    def __init__(self, path):
        self.path = path
        self.pidfile = None

    def __enter__(self):
        self.pidfile = open(self.path, "a+")
        try:
            fcntl.flock(self.pidfile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
        except IOError:
            raise SystemExit('Already running according to ' + self.path)
        self.pidfile.seek(0)
        self.pidfile.truncate()
        self.pidfile.write(str(os.getpid()))
        self.pidfile.flush()
        self.pidfile.seek(0)
        return self.pidfile

    def __exit__(self, exc_type=None, exc_value=None, exc_tb=None):
        try:
            self.pidfile.close()
        except IOError as err:
            # ok if file was just closed elsewhere
            if err.errno != 9:
                raise
        os.remove(self.path)


import hashlib
import logging
import random
import socketio
import time
import traceback
from cookies import Cookie, Cookies
from gevent.timeout import Timeout
from gevent.queue import Queue
from socketio.namespace import BaseNamespace
from socketio.mixins import RoomsMixin, BroadcastMixin

from aj.api import *
from aj.api.http import *
from aj.gate.session import Session
from aj.gate.worker import WorkerError
from aj.util import str_fsize


class SocketIONamespace (BaseNamespace, RoomsMixin, BroadcastMixin):
    name = '/socket'
    """ Endpoint ID """

    def __init__(self, context, env, *args, **kwargs):
        BaseNamespace.__init__(self, env, *args, **kwargs)
        self.context = context

        session = GateMiddleware.get(self.context).obtain_session(env)
        self.gate = session.gate
        self.spawn(self._stream_reader)

    def _stream_reader(self):
        q = self.gate.q_socket_messages.register()
        while True:
            resp = q.get()
            self.emit('message', resp.object['message'])

    def _send_worker_event(self, event, message=None):
        self.gate.stream.send({
            'type': 'socket',
            'event': event,
            'namespace': id(self),
            'message': message,
        })

    def recv_connect(self):
        logging.debug('Socket %s connected' % id(self))
        self._send_worker_event('connect')

    def recv_disconnect(self):
        logging.debug('Socket %s disconnected' % id(self))
        self._send_worker_event('disconnect')
        self.disconnect(silent=True)

    def on_message(self, message, *args):
        logging.debug('Socket %s message: %s' % (id(self), repr(message)))
        self._send_worker_event('message', message)


@service
class SocketIORouteHandler (BaseHttpHandler):
    def __init__(self, context):
        self.namespaces = {
            '/socket': lambda *args, **kwargs: SocketIONamespace(context, *args, **kwargs)
        }

    def handle(self, context):
        return str(socketio.socketio_manage(context.env, self.namespaces, context))


@service
class GateMiddleware (object):
    def __init__(self, context):
        self.context = context
        self.sessions = {}

    def generate_session_key(self, env):
        hash = str(random.random())
        hash += env.get('REMOTE_ADDR', '')
        hash += env.get('HTTP_USER_AGENT', '')
        hash += env.get('HTTP_HOST', '')
        return hashlib.sha1(hash).hexdigest()

    def vacuum(self):
        """
        Eliminates dead sessions
        """
        for session in [x for x in self.sessions.values() if x.is_dead()]:
            session.destroy()
            del self.sessions[session.key]

    def destroy(self):
        for session in self.sessions.values():
            session.deactivate()
        self.vacuum()

    def open_session(self, env):
        """
        Creates a new session for the :class:`aj.http.HttpContext`
        """
        client_info = {
            'address': env.get('REMOTE_ADDR', None),
        }
        session_key = self.generate_session_key(env)
        session = Session(session_key, client_info=client_info)
        self.sessions[session_key] = session
        return session

    def obtain_session(self, env):
        cookie_str = env.get('HTTP_COOKIE', None)
        session = None
        if cookie_str:
            cookie = Cookies.from_request(
                cookie_str,
                ignore_bad_cookies=True,
            ).get('session', None)
            if cookie and cookie.value:
                if cookie.value in self.sessions:
                    # Session found
                    session = self.sessions[cookie.value]
                    if session.is_dead():
                        session = None
        if session is None:
            session = self.open_session(env)
        return session

    def handle(self, http_context):
        #logging.debug('Incoming: %s' % http_context.path)
        start_time = time.time()

        self.vacuum()
        session = self.obtain_session(http_context.env)
        session.set_cookie(http_context)
        session.touch()

        if http_context.path.startswith('/socket'):
            #return http_context.respond_forbidden()
            http_context.fallthrough(SocketIORouteHandler.get(self.context))
            http_context.respond_ok()
            return 'Socket closed'

        request_object = {
            'type': 'http',
            'context': http_context.serialize(),
        }

        try:
            timeout = 30
            with Timeout(timeout) as t:
                q = session.gate.q_http_replies.register()
                rq = session.gate.stream.send(request_object)
                while True:
                    resp = q.get(t)
                    if resp.id == rq.id:
                        break
        except Timeout:
            http_context.respond('504 Gateway Timeout')
            return 'Worker timeout'

        if 'error' in resp.object:
            raise WorkerError(resp.object)

        for header in resp.object['headers']:
            http_context.add_header(*header)
        http_context.respond(resp.object['status'])
        content = resp.object['content']

        end_time = time.time()
        logging.debug('%.03fs %12s   %s %s %s' % (
            end_time - start_time,
            str_fsize(len(content[0] if content else [])),
            str(http_context.status).split()[0],
            http_context.env['REQUEST_METHOD'],
            http_context.path
        ))
        return content
